/** Huffman coding tree Main program.
    First, read from the data file a set of strings and associated
    frequencies.  These are placed onto a list of (single node)
    Huffman trees.  Next, build a single Huffman coding tree for
    the set.  The strings and their codes are then output, with
    CodeTable storing the coding for each input string.  Next, read
    commands to "encode" or "decode" strings, providing the
    appropriate output.

    This distribution should come with three test files: huffex.dat,
    hufftest.dat, and huffbook.dat. You can use any of these to test the
    program by reading them in from standard input. For example (in
    Linux) type to the command prompt:

    java Huffman < huffbook.dat
*/

import java.io.*;
import java.util.*;

public class Huffman {

private static Vector<Code> codeTable = new Vector<Code>(); // The code lookup table
private static int total = 0;  // Total weight of codes

/** Read a list of strings and frequencies,
    creating initial list of Huffman coding tree nodes */
static void readFreqs(BufferedReader d, List<HuffTree> hufflist)
     throws IOException {
  String line;
  char letter;
  int freq;
  HuffTree temptree;

  while((line = d.readLine()) != null) {
    // process the entry, creating a new HuffTree
    if (line.substring(0, 3).equals("end"))
      return; // Done with frequencies

	letter = line.charAt(0);
	freq = Integer.parseInt(line.substring(2));
    temptree = new HuffTree(new LettFreq(freq, letter));

    // Put in the list in sorted order
    // WARNING: This may be considered buggy.  Nodes with equal
	// weight will be put in reverse order of appearance, not
	// in alphabetical order.
    for (hufflist.moveToStart();
         hufflist.currPos() < hufflist.length(); hufflist.next())
      if (temptree.weight() <= hufflist.getValue().weight()) {
	    hufflist.insert(temptree);
	    break;
	  }
    if (hufflist.currPos()>= hufflist.length())
      hufflist.append(temptree);
  }
}
// Build a Huffman tree from list hufflist
static HuffTree buildTree(List<HuffTree> hufflist) {
  HuffTree tmp1, tmp2, tmp3;
  LettFreq tmpnode;

  for(hufflist.moveToPos(1); hufflist.length() > 1;
      hufflist.moveToPos(1)) {
    // While at least two items left
    hufflist.moveToStart();
    tmp1 = hufflist.remove();
    tmp2 = hufflist.remove();
    tmpnode = new LettFreq(tmp1.weight() + tmp2.weight());
    tmp3 = new HuffTree(tmpnode, tmp1, tmp2);

    // return to the list in sorted order
    for (hufflist.moveToStart();
         hufflist.currPos() < hufflist.length();
         hufflist.next())
      if (tmp3.weight() <= hufflist.getValue().weight())
        { hufflist.insert(tmp3); break; } // Put in list
    if (hufflist.currPos() >= hufflist.length())
      hufflist.append(tmp3); // This is heaviest value
  }
  hufflist.moveToStart();    // This is only tree on list
  return hufflist.remove();  // Return the tree
}

// Find the index in CodeTable for the given codeword
static int getindex(char codeword)
{
  int i;

  for (i=0; i<codeTable.size(); i++)
    if (codeword == codeTable.elementAt(i).data())
      return i;
  return i; // Not found
}


// Print out the codes; insert these codes into CodeTable
static void outputTree(BinNode<LettFreq> node, String prefix) {
  assert node != null : "Bad input tree";
  // This is a full binary tree so must not be null subtree
  if (node.isLeaf()) {
    System.out.println(node.element().letter() + "\t" + prefix);
    codeTable.addElement(new Code(node.element().letter(), prefix));
    total += prefix.length() * node.element().weight();
  }
  else {
    outputTree(node.left(), prefix + "0");
    outputTree(node.right(), prefix + "1");
  }
}


// Process Encode and Decode commands
static void doCommands(BufferedReader d, HuffTree tree)
      throws IOException {
  BinNode<LettFreq> temp;
  String line;
  int i;

  while ((line = d.readLine()) != null) {
    if (line.substring(0, 6).equals("decode")) {
      for (i=0; line.charAt(i) != '"'; i++);
      System.out.print("Decode " + line.charAt(i++));
      temp = tree.root();
      for (; line.charAt(i) != '"'; i++) { // Traverse the tree
        if (line.charAt(i) == '0')
          temp = temp.left();
        else if (line.charAt(i) == '1')
          temp = temp.right();
        else
          assert false : "Bad input: " + line;
        if (temp.isLeaf()) {
          System.out.print(temp.element().letter());
          temp = tree.root(); // reset at root
        }
      }
      System.out.println(line.charAt(i));
    }
    else if (line.substring(0, 6).equals("encode")) {
      for (i=0; line.charAt(i) != '"'; i++);
      System.out.print("Encode " + line.charAt(i++));
      for (; line.charAt(i) != '"'; i++) {
        // Assume codes are characters.  Should generalize this.
        int index = getindex(line.charAt(i));
        assert index < codeTable.size() :
		  "Tried to find code of bad character|" +
		  line.charAt(i) + "|";
        System.out.print(codeTable.elementAt(index).code());
      }
      System.out.println(line.charAt(i));
    }
    else
      assert false : "Bad command line: " + line;
  } // while
}

public static void main(String argv[])
       throws FileNotFoundException, IOException {
  List<HuffTree> templist = new LList<HuffTree>();
  BufferedReader f;

  if (argv.length == 0) // Read from standard input
      f = new BufferedReader(new InputStreamReader(System.in));
  else // Data file entered as parameter
      f = new BufferedReader(new InputStreamReader(new FileInputStream(argv[0])));
  System.out.println("Read frequencies");
  readFreqs(f, templist);
  System.out.println("Build the tree");
  HuffTree codes = buildTree(templist);
  System.out.println("Output the tree");
  outputTree(codes.root(), "");
  System.out.println("Average code length is " +
		(double)total/(double)codes.weight());
  doCommands(f, codes);
  System.in.read();
}

}
